# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

# This set of string-manipulation routines is strongly inspired by Arrow:

# string transforms
# https://arrow.apache.org/docs/python/api/compute.html#string-transforms
from __future__ import annotations
from awkward.operations.str.akstr_capitalize import *

# string padding
# https://arrow.apache.org/docs/python/api/compute.html#string-padding
from awkward.operations.str.akstr_center import *

# containment tests
# https://arrow.apache.org/docs/python/api/compute.html#containment-tests
from awkward.operations.str.akstr_count_substring import *
from awkward.operations.str.akstr_count_substring_regex import *
from awkward.operations.str.akstr_ends_with import *

# string component extraction
# https://arrow.apache.org/docs/python/api/compute.html#string-component-extraction
from awkward.operations.str.akstr_extract_regex import *
from awkward.operations.str.akstr_find_substring import *
from awkward.operations.str.akstr_find_substring_regex import *
from awkward.operations.str.akstr_index_in import *

# string predicates
# https://arrow.apache.org/docs/python/api/compute.html#string-predicates
from awkward.operations.str.akstr_is_alnum import *
from awkward.operations.str.akstr_is_alpha import *
from awkward.operations.str.akstr_is_ascii import *
from awkward.operations.str.akstr_is_decimal import *
from awkward.operations.str.akstr_is_digit import *
from awkward.operations.str.akstr_is_in import *
from awkward.operations.str.akstr_is_lower import *
from awkward.operations.str.akstr_is_numeric import *
from awkward.operations.str.akstr_is_printable import *
from awkward.operations.str.akstr_is_space import *
from awkward.operations.str.akstr_is_title import *
from awkward.operations.str.akstr_is_upper import *

# string joining
# https://arrow.apache.org/docs/python/api/compute.html#string-joining
from awkward.operations.str.akstr_join import *
from awkward.operations.str.akstr_join_element_wise import *
from awkward.operations.str.akstr_length import *
from awkward.operations.str.akstr_lower import *
from awkward.operations.str.akstr_lpad import *

# string trimming
# https://arrow.apache.org/docs/python/api/compute.html#string-trimming
from awkward.operations.str.akstr_ltrim import *
from awkward.operations.str.akstr_ltrim_whitespace import *
from awkward.operations.str.akstr_match_like import *
from awkward.operations.str.akstr_match_substring import *
from awkward.operations.str.akstr_match_substring_regex import *
from awkward.operations.str.akstr_repeat import *
from awkward.operations.str.akstr_replace_slice import *
from awkward.operations.str.akstr_replace_substring import *
from awkward.operations.str.akstr_replace_substring_regex import *
from awkward.operations.str.akstr_reverse import *
from awkward.operations.str.akstr_rpad import *
from awkward.operations.str.akstr_rtrim import *
from awkward.operations.str.akstr_rtrim_whitespace import *

# string slicing
# https://arrow.apache.org/docs/python/api/compute.html#string-slicing
from awkward.operations.str.akstr_slice import *
from awkward.operations.str.akstr_split_pattern import *
from awkward.operations.str.akstr_split_pattern_regex import *

# string splitting
# https://arrow.apache.org/docs/python/api/compute.html#string-splitting
from awkward.operations.str.akstr_split_whitespace import *
from awkward.operations.str.akstr_starts_with import *
from awkward.operations.str.akstr_swapcase import *
from awkward.operations.str.akstr_title import *

# dictionary-encoding (exclusively for strings)
# https://arrow.apache.org/docs/python/api/compute.html#associative-transforms
from awkward.operations.str.akstr_to_categorical import *
from awkward.operations.str.akstr_trim import *
from awkward.operations.str.akstr_trim_whitespace import *
from awkward.operations.str.akstr_upper import *


def _drop_option_preserving_form(layout, ensure_empty_mask: bool = False):
    from awkward._do import recursively_apply
    from awkward.contents import UnmaskedArray, IndexedOptionArray, IndexedArray

    def action(_, continuation, **kwargs):
        this = continuation()

        if not this.is_option:
            return this
        elif isinstance(this, UnmaskedArray):
            return this.content
        else:
            index_nplike = this.backend.index_nplike
            assert not (
                ensure_empty_mask
                and index_nplike.known_data
                and index_nplike.any(this.mask_as_bool(valid_when=False))
            ), "did not expect option type, but arrow returned a non-erasable option"
            # Re-write indexed options as indexed
            if isinstance(this, IndexedOptionArray):
                return IndexedArray(
                    this.index, this.content, parameters=this._parameters
                )
            else:
                return this.content

    return recursively_apply(layout, action, return_simplified=False)


def _apply_through_arrow(
    operation,
    /,
    *args,
    generate_bitmasks=False,
    expect_option_type=False,
    string_to32=False,
    bytestring_to32=False,
    ensure_empty_mask=False,
    **kwargs,
):
    from awkward._backends.dispatch import backend_of
    from awkward.contents.content import Content
    from awkward.operations.ak_from_arrow import from_arrow
    from awkward.operations.ak_to_arrow import to_arrow
    from awkward._backends.typetracer import TypeTracerBackend

    backend = backend_of(*args)

    typetracer = TypeTracerBackend.instance()
    if backend is typetracer:
        converted_args = [
            to_arrow(
                x.form.length_zero_array(highlevel=False),
                extensionarray=False,
                string_to32=string_to32,
                bytestring_to32=bytestring_to32,
            )
            if isinstance(x, Content)
            else x
            for x in args
        ]
        out = from_arrow(
            operation(*converted_args, **kwargs),
            generate_bitmasks=generate_bitmasks,
            highlevel=False,
        ).to_typetracer(forget_length=True)

    else:
        converted_args = [
            to_arrow(
                x,
                extensionarray=False,
                string_to32=string_to32,
                bytestring_to32=bytestring_to32,
            )
            if isinstance(x, Content)
            else x
            for x in args
        ]
        out = from_arrow(
            operation(*converted_args, **kwargs),
            generate_bitmasks=generate_bitmasks,
            highlevel=False,
        )

    if expect_option_type:
        return out
    else:
        return _drop_option_preserving_form(out, ensure_empty_mask=ensure_empty_mask)


def _get_ufunc_action(
    utf8_function,
    ascii_function,
    *args,
    generate_bitmasks=False,
    bytestring_to_string=False,
    expect_option_type=False,
    **kwargs,
):
    def action(layout, **absorb):
        if layout.is_list and layout.parameter("__array__") == "string":
            return _apply_through_arrow(
                utf8_function,
                layout,
                *args,
                generate_bitmasks=generate_bitmasks,
                expect_option_type=expect_option_type,
                **kwargs,
            )

        elif layout.is_list and layout.parameter("__array__") == "bytestring":
            if bytestring_to_string:
                out = _apply_through_arrow(
                    ascii_function,
                    layout.copy(
                        content=layout.content.copy(parameters={"__array__": "char"}),
                        parameters={"__array__": "string"},
                    ),
                    *args,
                    generate_bitmasks=generate_bitmasks,
                    expect_option_type=expect_option_type,
                    **kwargs,
                )
                if out.is_list and out.parameter("__array__") == "string":
                    out = out.copy(
                        content=out.content.copy(parameters={"__array__": "byte"}),
                        parameters={"__array__": "bytestring"},
                    )
                return out

            else:
                return _apply_through_arrow(
                    ascii_function,
                    layout,
                    *args,
                    generate_bitmasks=generate_bitmasks,
                    expect_option_type=expect_option_type,
                    **kwargs,
                )

    return action


def _get_split_action(
    utf8_function,
    ascii_function,
    *args,
    generate_bitmasks=False,
    bytestring_to_string=False,
    **kwargs,
):
    from awkward._backends.typetracer import TypeTracerBackend
    from awkward.forms import ListOffsetForm, NumpyForm

    typetracer = TypeTracerBackend.instance()

    # FIXME: this workaround for typetracer is required because
    #        split_XXX does not support length-zero arrays
    #        c.f. https://github.com/apache/arrow/issues/37437
    def action(layout, **_):
        if layout.backend is typetracer:
            if layout.is_list and layout.parameter("__array__") == "string":
                return (
                    ListOffsetForm(
                        "i32",
                        ListOffsetForm(
                            layout.form.offsets,
                            NumpyForm("uint8", parameters={"__array__": "char"}),
                            parameters={"__array__": "string"},
                        ),
                    )
                    .length_zero_array(highlevel=False)
                    .to_typetracer(forget_length=True)
                )

            elif layout.is_list and layout.parameter("__array__") == "bytestring":
                return (
                    ListOffsetForm(
                        "i32",
                        ListOffsetForm(
                            layout.form.offsets,
                            NumpyForm("uint8", parameters={"__array__": "byte"}),
                            parameters={"__array__": "bytestring"},
                        ),
                    )
                    .length_zero_array(highlevel=False)
                    .to_typetracer(forget_length=True)
                )
        else:
            if layout.is_list and layout.parameter("__array__") == "string":
                return _drop_option_preserving_form(
                    _apply_through_arrow(
                        utf8_function,
                        layout,
                        *args,
                        generate_bitmasks=generate_bitmasks,
                        **kwargs,
                    )
                )

            elif layout.is_list and layout.parameter("__array__") == "bytestring":
                if bytestring_to_string:
                    out = _drop_option_preserving_form(
                        _apply_through_arrow(
                            ascii_function,
                            layout.copy(
                                content=layout.content.copy(
                                    parameters={"__array__": "char"}
                                ),
                                parameters={"__array__": "string"},
                            ),
                            *args,
                            generate_bitmasks=generate_bitmasks,
                            **kwargs,
                        )
                    )
                    assert out.is_list

                    assert (
                        out.content.is_list
                        and out.content.parameter("__array__") == "string"
                    )
                    return out.copy(
                        content=out.content.copy(
                            content=out.content.content.copy(
                                parameters={"__array__": "byte"}
                            ),
                            parameters={"__array__": "bytestring"},
                        ),
                    )

                else:
                    return _drop_option_preserving_form(
                        _apply_through_arrow(
                            ascii_function,
                            layout,
                            *args,
                            generate_bitmasks=generate_bitmasks,
                            **kwargs,
                        )
                    )

    return action



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import contextlib

import numpy as np
import packaging.version
import pytest

import awkward as ak

NUMPY_HAS_NEP_50 = packaging.version.parse(np.__version__) >= packaging.version.Version(
    "1.24.0"
)


@pytest.mark.skipif(not NUMPY_HAS_NEP_50, reason="NEP-50 requires NumPy >= 1.24.0")
@pytest.mark.parametrize("backend", ["cpu", "typetracer"])
def test_with_nep_50(backend):
    array = ak.to_backend(np.arange(255, dtype=np.uint8), backend)
    assert array.layout.dtype == np.dtype(np.uint8)

    typed_scalar = np.uint64(0)
    assert (array + typed_scalar).layout.dtype == np.dtype(np.uint64)

    # With NEP-50, we can ask NumPy to use value-less type resolution
    warn_context = (
        pytest.warns(DeprecationWarning, match="out-of-bound Python integers")
        if backend == "cpu"
        else contextlib.nullcontext()
    )
    with warn_context:
        untyped_scalar = 512
        assert (array + untyped_scalar).layout.dtype == np.dtype(np.uint8)


@pytest.mark.skipif(NUMPY_HAS_NEP_50, reason="NumPy >= 1.24.0 has NEP-50 support")
@pytest.mark.parametrize("backend", ["cpu", "typetracer"])
def test_without_nep_50(backend):
    array = ak.to_backend(np.arange(255, dtype=np.uint8), backend)
    assert array.layout.dtype == np.dtype(np.uint8)

    # Without NEP-50, we still don't drop type information for typed-scalars,
    # unlike NumPy.
    typed_scalar = np.uint64(0)
    assert (array + typed_scalar).layout.dtype == np.dtype(np.uint64)

    # But, with untyped scalars, we're forced to rely on NumPy's ufunc loop resolution
    untyped_scalar = 512
    assert (array + untyped_scalar).layout.dtype == np.dtype(np.uint16)



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_IndexedArray32_getitem_nextcarry_outindex_64_1():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    toindex = cupy.array([123, 123, 123], dtype=cupy.int32)
    fromindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    lenindex = 3
    lencontent = 2
    funcC = cupy_backend['awkward_IndexedArray_getitem_nextcarry_outindex', cupy.int64, cupy.int32, cupy.int32]
    funcC(tocarry, toindex, fromindex, lenindex, lencontent)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [1, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_toindex = [0.0, 1.0, 2.0]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArray32_getitem_nextcarry_outindex_64_2():
    tocarry = cupy.array([123], dtype=cupy.int64)
    toindex = cupy.array([123], dtype=cupy.int32)
    fromindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    lenindex = 3
    lencontent = 2
    funcC = cupy_backend['awkward_IndexedArray_getitem_nextcarry_outindex', cupy.int64, cupy.int32, cupy.int32]

def test_cuda_awkward_IndexedArray32_getitem_nextcarry_outindex_64_3():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    toindex = cupy.array([123, 123, 123], dtype=cupy.int32)
    fromindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    lenindex = 3
    lencontent = 5
    funcC = cupy_backend['awkward_IndexedArray_getitem_nextcarry_outindex', cupy.int64, cupy.int32, cupy.int32]
    funcC(tocarry, toindex, fromindex, lenindex, lencontent)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 0, 0]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_toindex = [0.0, 1.0, 2.0]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import traceback

import cupy as cp
import numpy as np  # noqa: F401
import pytest

import awkward as ak
import awkward._connect.cuda

try:
    ak.numba.register_and_check()
except ImportError:
    pytest.skip(reason="too old Numba version", allow_module_level=True)


def test():
    v2_array = ak.Array([1, 2, 3, 4, 5]).layout

    stream = cp.cuda.Stream(non_blocking=True)

    v2_array_cuda = ak.to_backend(v2_array, "cuda")
    with cp.cuda.Stream() as stream:
        v2_array_cuda[10,]
        v2_array_cuda[11,]
        v2_array_cuda[12,]

    with pytest.raises(ValueError) as err:
        awkward._connect.cuda.synchronize_cuda(stream)

    assert isinstance(err.value, ValueError)

    message = "".join(traceback.format_exception(err.type, err.value, err.tb))
    assert (
        "ValueError: index out of range in compiled CUDA code "
        "(awkward_RegularArray_getitem_next_at)\n"
        "\n"
        "This error occurred while attempting to slice\n"
        "\n"
        "    <Array [1, 2, 3, 4, 5] type='5 * int64'>\n"
        "\n"
        "with\n"
        "\n"
        "    (10)\n"
    ) in message



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._backends.typetracer import TypeTracerBackend
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext, ensure_same_backend

__all__ = ("join_element_wise",)

typetracer = TypeTracerBackend.instance()


@high_level_function(module="ak.str")
def join_element_wise(*arrays, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        arrays: Array-like data (anything #ak.to_layout recognizes).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Broadcasts and concatenates all but the last array of strings in `arrays`;
    the last is used as a separator.

    Note: this function does not raise an error if the `array` does not
    contain any string or bytestring data.

    Requires the pyarrow library and calls
    [pyarrow.compute.binary_join_element_wise](https://arrow.apache.org/docs/python/generated/pyarrow.compute.binary_join_element_wise.html).

    Unlike Arrow's `binary_join_element_wise`, this function has no `null_handling`
    and `null_replacement` arguments. This function's behavior is like
    `null_handling="emit_null"` (Arrow's default). The other cases can be implemented
    with Awkward slices, #ak.drop_none, and #ak.fill_none.

    See also: #ak.str.join.
    """
    # Dispatch
    yield arrays

    # Implementation
    return _impl(arrays, highlevel, behavior, attrs)


def _impl(arrays, highlevel, behavior, attrs):
    from awkward._connect.pyarrow import import_pyarrow_compute
    from awkward.operations.str import _apply_through_arrow

    pc = import_pyarrow_compute("ak.str.join_element_wise")

    if len(arrays) < 1:
        raise TypeError("at least one array is required")

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layouts = ensure_same_backend(*(ctx.unwrap(x) for x in arrays))

    def action(layouts, **kwargs):
        if all(
            x.is_list and x.parameter("__array__") in ("string", "bytestring")
            for x in layouts
        ):
            return (_apply_through_arrow(pc.binary_join_element_wise, *layouts),)

    (out,) = ak._broadcasting.broadcast_and_apply(layouts, action)

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext, maybe_posaxis
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._regularize import regularize_axis
from awkward.errors import AxisError

__all__ = ("to_regular",)

np = NumpyMetadata.instance()


@high_level_function()
def to_regular(array, axis=1, *, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (int or None): The dimension at which this operation is applied.
            The outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc. If None, convert all
            variable dimensions into regular ones or raise a ValueError if that
            is not possible.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Converts a variable-length axis into a regular one, if possible.

        >>> irregular = ak.from_iter(np.arange(2*3*5).reshape(2, 3, 5))
        >>> irregular.type.show()
        2 * var * var * int64
        >>> ak.to_regular(irregular).type.show()
        2 * 3 * var * int64
        >>> ak.to_regular(irregular, axis=2).type.show()
        2 * var * 5 * int64
        >>> ak.to_regular(irregular, axis=-1).type.show()
        2 * var * 5 * int64

    But truly irregular data cannot be converted.

        >>> ak.to_regular(ak.Array([[1, 2, 3], [], [4, 5]]))
        ValueError: while calling
            ak.to_regular(
                array = <Array [[1, 2, 3], [], [4, 5]] type='3 * var * int64'>
                axis = 1
                highlevel = True
                behavior = None
            )
        Error details: cannot convert to RegularArray because subarray lengths are not regular

    See also #ak.from_regular.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, highlevel, behavior, attrs)


def _impl(array, axis, highlevel, behavior, attrs):
    axis = regularize_axis(axis)
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=False, primitive_policy="error")

    if axis is None:

        def action(layout, continuation, **kwargs):
            if layout.is_list:
                return continuation().to_RegularArray()

        out = ak._do.recursively_apply(layout, action)

    elif maybe_posaxis(layout, axis, 1) == 0:
        out = layout  # the top-level can only be regular (ArrayType)

    else:

        def action(layout, depth, **kwargs):
            posaxis = maybe_posaxis(layout, axis, depth)
            if posaxis == depth and layout.is_list:
                return layout.to_RegularArray()

            elif layout.is_leaf:
                raise AxisError(
                    f"axis={axis} exceeds the depth of this array ({depth})"
                )

        out = ak._do.recursively_apply(layout, action)

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numbers

import awkward as ak
from awkward._backends.typetracer import TypeTracerBackend
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext, ensure_same_backend
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("repeat",)

typetracer = TypeTracerBackend.instance()
np = NumpyMetadata.instance()


@high_level_function(module="ak.str")
def repeat(array, num_repeats, *, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        num_repeats: (int, or an array of them to broadcast): number of times
            to repeat each element
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Replaces any string-valued or bytestring-valued data with the same value
    repeated `num_repeats` times, which can be a scalar integer or a
    (broadcasted) array of integers.

    Note: this function does not raise an error if the `array` does not
    contain any string or bytestring data.

    Requires the pyarrow library and calls
    [pyarrow.compute.binary_repeat](https://arrow.apache.org/docs/python/generated/pyarrow.compute.binary_repeat.html)
    or
    [pyarrow.compute.binary_repeat](https://arrow.apache.org/docs/python/generated/pyarrow.compute.binary_repeat.html)
    on strings and bytestrings, respectively.
    """
    # Dispatch
    yield array, num_repeats

    # Implementation
    return _impl(array, num_repeats, highlevel, behavior, attrs)


def _impl(array, num_repeats, highlevel, behavior, attrs):
    from awkward._connect.pyarrow import import_pyarrow_compute
    from awkward.operations.str import _apply_through_arrow

    pc = import_pyarrow_compute("ak.str.repeat")

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout, num_repeats_layout = ensure_same_backend(
            ctx.unwrap(array, allow_record=False),
            ctx.unwrap(
                num_repeats, allow_record=False, primitive_policy="pass-through"
            ),
        )
    if isinstance(num_repeats_layout, ak.contents.Content):

        def action(inputs, **kwargs):
            if inputs[0].is_list and inputs[0].parameter("__array__") in (
                "string",
                "bytestring",
            ):
                if not (
                    inputs[1].is_numpy and np.issubdtype(inputs[1].dtype, np.integer)
                ):
                    raise TypeError(
                        "num_repeats must be an integer or broadcastable to integers"
                    )

                return (_apply_through_arrow(pc.binary_repeat, *inputs),)

        (out,) = ak._broadcasting.broadcast_and_apply(
            (layout, num_repeats_layout), action
        )

    else:
        if not isinstance(num_repeats, numbers.Integral):
            raise TypeError(
                "num_repeats must be an integer or broadcastable to integers"
            )

        def action(layout, **kwargs):
            if layout.is_list and layout.parameter("__array__") in (
                "string",
                "bytestring",
            ):
                return _apply_through_arrow(pc.binary_repeat, layout, num_repeats)

        out = ak._do.recursively_apply(layout, action)

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext, ensure_same_backend
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("isclose",)

np = NumpyMetadata.instance()


@high_level_function()
def isclose(
    a,
    b,
    rtol=1e-05,
    atol=1e-08,
    equal_nan=False,
    *,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        a: Array-like data (anything #ak.to_layout recognizes).
        b: Array-like data (anything #ak.to_layout recognizes).
        rtol (float): The relative tolerance parameter.
        atol (float): The absolute tolerance parameter.
        equal_nan (bool): Whether to compare `NaN` as equal. If True, `NaN` in
            `a` will be considered equal to `NaN` in `b`.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Implements [np.isclose](https://numpy.org/doc/stable/reference/generated/numpy.isclose.html)
    for Awkward Arrays.
    """
    # Dispatch
    yield a, b

    # Implementation
    return _impl(a, b, rtol, atol, equal_nan, highlevel, behavior, attrs)


def _impl(a, b, rtol, atol, equal_nan, highlevel, behavior, attrs):
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layouts = ensure_same_backend(
            ctx.unwrap(a, allow_record=False),
            ctx.unwrap(b, allow_record=False),
        )

    def action(inputs, backend, **kwargs):
        if all(isinstance(x, ak.contents.NumpyArray) for x in inputs):
            return (
                ak.contents.NumpyArray(
                    backend.nplike.isclose(
                        inputs[0]._raw(backend.nplike),
                        inputs[1]._raw(backend.nplike),
                        rtol=rtol,
                        atol=atol,
                        equal_nan=equal_nan,
                    )
                ),
            )

    out = ak._broadcasting.broadcast_and_apply(layouts, action)
    assert isinstance(out, tuple) and len(out) == 1

    return ctx.wrap(out[0], highlevel=highlevel)


@ak._connect.numpy.implements("isclose")
def _nep_18_impl(a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
    return isclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test():
    layout = ak.to_layout([{"x": 10}]).to_typetracer(forget_length=True)
    named = ak.with_name(layout, "x_like")
    assert ak.parameters(named) == {"__record__": "x_like"}



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


class ArrayBehavior(ak.Array):
    def impl(self):
        return True


class RecordBehavior(ak.Record):
    def impl(self):
        return True


BEHAVIOR = {("*", "impl"): ArrayBehavior, "impl": RecordBehavior}


def test_array_layout():
    array = ak.Array([{"x": 1}, {"y": 3}], behavior=BEHAVIOR)
    assert not isinstance(array, ArrayBehavior)

    array.layout = ak.with_name([{"x": 1}, {"y": 3}], "impl", highlevel=False)
    assert isinstance(array, ArrayBehavior)
    assert array.impl()


def test_array_behavior():
    array = ak.Array([{"x": 1}, {"y": 3}], with_name="impl")
    assert not isinstance(array, ArrayBehavior)

    array.behavior = BEHAVIOR
    assert isinstance(array, ArrayBehavior)
    assert array.impl()


def test_record_layout():
    record = ak.Record({"x": 1}, behavior=BEHAVIOR)
    assert not isinstance(record, RecordBehavior)

    record.layout = ak.with_name({"x": 1}, "impl", highlevel=False)
    assert isinstance(record, RecordBehavior)
    assert record.impl()


def test_record_behavior():
    record = ak.Record({"x": 1}, with_name="impl")
    assert not isinstance(record, RecordBehavior)

    record.behavior = BEHAVIOR
    assert isinstance(record, RecordBehavior)
    assert record.impl()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_union_simplification():
    array = ak.Array(
        ak.contents.UnionArray(
            ak.index.Index8(np.arange(64, dtype=np.int8) % 2),
            ak.index.Index64(np.arange(64, dtype=np.int64) // 2),
            [
                ak.contents.RecordArray(
                    [ak.contents.NumpyArray(np.arange(64, dtype=np.int64))], ["x"]
                ),
                ak.contents.RecordArray(
                    [
                        ak.contents.NumpyArray(np.arange(64, dtype=np.int64)),
                        ak.contents.NumpyArray(np.arange(64, dtype=np.int8)),
                    ],
                    ["x", "y"],
                ),
            ],
        )
    )

    form, length, container = ak.to_buffers(array)

    assert form.to_dict() == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i64",
        "contents": [
            {
                "class": "RecordArray",
                "fields": ["x"],
                "contents": [
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "node2",
                    }
                ],
                "parameters": {},
                "form_key": "node1",
            },
            {
                "class": "RecordArray",
                "fields": ["x", "y"],
                "contents": [
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "node4",
                    },
                    {
                        "class": "NumpyArray",
                        "primitive": "int8",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "node5",
                    },
                ],
                "parameters": {},
                "form_key": "node3",
            },
        ],
        "parameters": {},
        "form_key": "node0",
    }

    projected_form = {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i64",
        "contents": [
            {
                "class": "RecordArray",
                "fields": ["x"],
                "contents": [
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "node2",
                    }
                ],
                "parameters": {},
                "form_key": "node1",
            },
            {
                "class": "RecordArray",
                "fields": ["x"],
                "contents": [
                    {
                        "class": "NumpyArray",
                        "primitive": "int64",
                        "inner_shape": [],
                        "parameters": {},
                        "form_key": "node4",
                    }
                ],
                "parameters": {},
                "form_key": "node3",
            },
        ],
        "parameters": {},
        "form_key": "node0",
    }
    container.pop("node5-data")
    projected = ak.from_buffers(
        projected_form, length, container, allow_noncanonical_form=True
    )
    assert projected.layout.form.to_dict(verbose=False) == {
        "class": "IndexedArray",
        "index": "i64",
        "content": {"class": "RecordArray", "fields": ["x"], "contents": ["int64"]},
    }
    assert ak.almost_equal(array[["x"]], projected)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.placeholder import PlaceholderArray
from awkward._nplikes.shape import unknown_length

numpy = Numpy.instance()


def test_numpyarray():
    layout = ak.from_buffers(
        {"class": "NumpyArray", "primitive": "int64", "form_key": "node0"},
        10,
        {"node0-data": PlaceholderArray(numpy, (10,), np.int64)},
        highlevel=False,
    )
    assert layout.length == 10

    # Content too small
    with pytest.raises(TypeError, match=r"is less than size of form"):
        ak.from_buffers(
            {"class": "NumpyArray", "primitive": "int64", "form_key": "node0"},
            10,
            {"node0-data": PlaceholderArray(numpy, (9,), np.int64)},
            highlevel=False,
        )

    # Unknown length content at top-level
    layout = ak.from_buffers(
        {"class": "NumpyArray", "primitive": "int64", "form_key": "node0"},
        10,
        {"node0-data": PlaceholderArray(numpy, (unknown_length,), np.int64)},
        highlevel=False,
    )
    assert layout.length == 10


def test_listoffsetarray_numpyarray():
    # Unknown data
    layout = ak.from_buffers(
        {
            "class": "ListOffsetArray",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "offsets": "i64",
            "form_key": "node0",
        },
        2,
        {
            "node0-offsets": np.array([0, 1, 2], dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (10,), dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 2
    assert layout.content.length == 2

    # Unknown offsets above known data
    layout = ak.from_buffers(
        {
            "class": "ListOffsetArray",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "offsets": "i64",
            "form_key": "node0",
        },
        2,
        {
            "node0-offsets": PlaceholderArray(numpy, (3,), dtype=np.int64),
            "node1-data": np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 2
    assert layout.content.length is unknown_length

    # Unknown offsets and unknown data
    layout = ak.from_buffers(
        {
            "class": "ListOffsetArray",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "offsets": "i64",
            "form_key": "node0",
        },
        2,
        {
            "node0-offsets": PlaceholderArray(numpy, (3,), dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (10,), dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 2
    assert layout.content.length is unknown_length

    # Unknown offsets and unknown data with unknown offset length
    layout = ak.from_buffers(
        {
            "class": "ListOffsetArray",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "offsets": "i64",
            "form_key": "node0",
        },
        2,
        {
            "node0-offsets": PlaceholderArray(numpy, (unknown_length,), dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (10,), dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 2
    assert layout.content.length is unknown_length


def test_listarray_numpyarray():
    # Unknown data
    layout = ak.from_buffers(
        {
            "class": "ListArray",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "starts": "i64",
            "stops": "i64",
            "form_key": "node0",
        },
        2,
        {
            "node0-starts": np.array([0, 1], dtype=np.int64),
            "node0-stops": np.array([1, 2], dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (10,), dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 2
    assert layout.content.length == 2

    # Unknown offsets
    layout = ak.from_buffers(
        {
            "class": "ListArray",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "starts": "i64",
            "stops": "i64",
            "form_key": "node0",
        },
        2,
        {
            "node0-starts": PlaceholderArray(numpy, (2,), dtype=np.int64),
            "node0-stops": PlaceholderArray(numpy, (2,), dtype=np.int64),
            "node1-data": np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 2
    assert layout.content.length is unknown_length

    # Unknown offsets and unknown data
    layout = ak.from_buffers(
        {
            "class": "ListArray",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "starts": "i64",
            "stops": "i64",
            "form_key": "node0",
        },
        2,
        {
            "node0-starts": PlaceholderArray(numpy, (2,), dtype=np.int64),
            "node0-stops": PlaceholderArray(numpy, (2,), dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (10,), dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 2
    assert layout.content.length is unknown_length

    # Unknown starts, stops, and data with unknown starts lengths
    layout = ak.from_buffers(
        {
            "class": "ListArray",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "starts": "i64",
            "stops": "i64",
            "form_key": "node0",
        },
        2,
        {
            "node0-starts": PlaceholderArray(numpy, (unknown_length,), dtype=np.int64),
            "node0-stops": PlaceholderArray(numpy, (2,), dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (10,), dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 2
    assert layout.content.length is unknown_length


def test_indexedoptionarray():
    # Unknown data
    layout = ak.from_buffers(
        {
            "class": "IndexedOptionArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "form_key": "node0",
        },
        3,
        {
            "node0-index": np.array([0, 1, 2], dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (3,), np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.content.length == 3

    # Unknown index
    layout = ak.from_buffers(
        {
            "class": "IndexedOptionArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "form_key": "node0",
        },
        3,
        {
            "node0-index": PlaceholderArray(numpy, (3,), np.int64),
            "node1-data": np.array([0, 1, 2, 3, 4, 5], dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.content.length is unknown_length

    # Unknown index and data
    layout = ak.from_buffers(
        {
            "class": "IndexedOptionArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "form_key": "node0",
        },
        3,
        {
            "node0-index": PlaceholderArray(numpy, (3,), np.int64),
            "node1-data": PlaceholderArray(numpy, (6,), np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.content.length is unknown_length

    # Unknown index and data with unknown index length
    layout = ak.from_buffers(
        {
            "class": "IndexedOptionArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "form_key": "node0",
        },
        3,
        {
            "node0-index": PlaceholderArray(numpy, (unknown_length,), np.int64),
            "node1-data": PlaceholderArray(numpy, (6,), np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.content.length is unknown_length


def test_indexedarray():
    # Unknown data
    layout = ak.from_buffers(
        {
            "class": "IndexedArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "form_key": "node0",
        },
        3,
        {
            "node0-index": np.array([0, 1, 2], dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (3,), np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.content.length == 3

    # Unknown index
    layout = ak.from_buffers(
        {
            "class": "IndexedArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "form_key": "node0",
        },
        3,
        {
            "node0-index": PlaceholderArray(numpy, (3,), np.int64),
            "node1-data": np.array([0, 1, 2, 3, 4, 5], dtype=np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.content.length is unknown_length

    # Unknown data
    layout = ak.from_buffers(
        {
            "class": "IndexedArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "form_key": "node0",
        },
        3,
        {
            "node0-index": PlaceholderArray(numpy, (3,), np.int64),
            "node1-data": PlaceholderArray(numpy, (6,), np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.content.length is unknown_length

    # Unknown index and data, with unknown index length
    layout = ak.from_buffers(
        {
            "class": "IndexedArray",
            "index": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "form_key": "node1",
            },
            "form_key": "node0",
        },
        3,
        {
            "node0-index": PlaceholderArray(numpy, (unknown_length,), np.int64),
            "node1-data": PlaceholderArray(numpy, (6,), np.int64),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.content.length is unknown_length


def test_unionarray():
    # Unknown data
    layout = ak.from_buffers(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "i64",
            "contents": [
                {
                    "class": "NumpyArray",
                    "primitive": "int64",
                    "form_key": "node1",
                },
                {
                    "class": "NumpyArray",
                    "primitive": "datetime64[D]",
                    "form_key": "node2",
                },
            ],
            "form_key": "node0",
        },
        3,
        {
            "node0-tags": np.array([0, 0, 1], dtype=np.int8),
            "node0-index": np.array([0, 1, 0], dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (3,), np.int64),
            "node2-data": PlaceholderArray(numpy, (6,), np.dtype("datetime64[D]")),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.contents[0].length == 2
    assert layout.contents[1].length == 1

    # Unknown tags
    layout = ak.from_buffers(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "i64",
            "contents": [
                {
                    "class": "NumpyArray",
                    "primitive": "int64",
                    "form_key": "node1",
                },
                {
                    "class": "NumpyArray",
                    "primitive": "datetime64[D]",
                    "form_key": "node2",
                },
            ],
            "form_key": "node0",
        },
        3,
        {
            "node0-tags": PlaceholderArray(numpy, (3,), np.int8),
            "node0-index": np.array([0, 1, 0], dtype=np.int64),
            "node1-data": np.array([0, 1, 2], np.int64),
            "node2-data": np.array([0, 1, 2, 3, 4, 5, 6], np.dtype("datetime64[D]")),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.contents[0].length is unknown_length
    assert layout.contents[1].length is unknown_length

    # Unknown index
    layout = ak.from_buffers(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "i64",
            "contents": [
                {
                    "class": "NumpyArray",
                    "primitive": "int64",
                    "form_key": "node1",
                },
                {
                    "class": "NumpyArray",
                    "primitive": "datetime64[D]",
                    "form_key": "node2",
                },
            ],
            "form_key": "node0",
        },
        3,
        {
            "node0-tags": np.array([0, 0, 1], dtype=np.int8),
            "node0-index": PlaceholderArray(numpy, (3,), np.int64),
            "node1-data": np.array([0, 1, 2], np.int64),
            "node2-data": np.array([0, 1, 2, 3, 4, 5, 6], np.dtype("datetime64[D]")),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.contents[0].length is unknown_length
    assert layout.contents[1].length is unknown_length

    # Unknown content length
    layout = ak.from_buffers(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "i64",
            "contents": [
                {
                    "class": "NumpyArray",
                    "primitive": "int64",
                    "form_key": "node1",
                },
                {
                    "class": "NumpyArray",
                    "primitive": "datetime64[D]",
                    "form_key": "node2",
                },
            ],
            "form_key": "node0",
        },
        3,
        {
            "node0-tags": np.array([0, 0, 1], dtype=np.int8),
            "node0-index": np.array([0, 1, 0], dtype=np.int64),
            "node1-data": PlaceholderArray(numpy, (unknown_length,), np.int64),
            "node2-data": PlaceholderArray(numpy, (6,), np.dtype("datetime64[D]")),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.contents[0].length == 2
    assert layout.contents[1].length == 1

    # Unknown tags, index, and data
    layout = ak.from_buffers(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "i64",
            "contents": [
                {
                    "class": "NumpyArray",
                    "primitive": "int64",
                    "form_key": "node1",
                },
                {
                    "class": "NumpyArray",
                    "primitive": "datetime64[D]",
                    "form_key": "node2",
                },
            ],
            "form_key": "node0",
        },
        3,
        {
            "node0-tags": PlaceholderArray(numpy, (3,), np.int8),
            "node0-index": PlaceholderArray(numpy, (3,), np.int64),
            "node1-data": PlaceholderArray(numpy, (3,), np.int64),
            "node2-data": PlaceholderArray(numpy, (6,), np.dtype("datetime64[D]")),
        },
        highlevel=False,
    )
    assert layout.length == 3
    assert layout.contents[0].length is unknown_length
    assert layout.contents[1].length is unknown_length



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import packaging.version
import pytest

import awkward as ak

if packaging.version.Version(np.__version__) < packaging.version.Version("1.24.0"):
    pytest.skip("Only NumPy>=1.24 supports `resolve_dtypes`", allow_module_level=True)


def test_bool():
    array = ak.Array([[True, False], [True, False, False]])
    assert (array == False).to_list() == [[False, True], [False, True, True]]  # noqa: E712


def test_int():
    array = ak.Array([[1, 2], [3, 4, 5]])
    assert (array == 2).to_list() == [[False, True], [False, False, False]]


def test_float():
    array = ak.Array([[1.0, 2.7], [2.0, 7.0, 8.8]])
    assert (array == 1.0).to_list() == [[True, False], [False, False, False]]


def test_complex():
    array = ak.Array([[1.0j, 2.7], [2.0 + 1j, 7.0, 8.8]])
    assert (array == 1.0j).to_list() == [[True, False], [False, False, False]]


def test_datetime64():
    array = ak.Array([[np.datetime64(1, "D"), np.datetime64(10, "D")]])
    assert (array == np.datetime64(10, "D")).to_list() == [[False, True]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward_cpp.lib._ext import ArrayBuilder

__all__ = ["ArrayBuilder"]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from collections.abc import Callable

import awkward as ak
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._parameters import type_parameters_equal
from awkward._typing import DType, Iterator, JSONSerializable, Self, final
from awkward._util import UNSET
from awkward.forms.form import Form, _SpecifierMatcher, index_to_dtype

__all__ = ("ByteMaskedForm",)

np = NumpyMetadata.instance()


@final
class ByteMaskedForm(Form):
    is_option = True

    def __init__(
        self,
        mask,
        content,
        valid_when,
        *,
        parameters=None,
        form_key=None,
    ):
        if not isinstance(mask, str):
            raise TypeError(
                f"{type(self).__name__} 'mask' must be of type str, not {mask!r}"
            )
        if not isinstance(content, Form):
            raise TypeError(
                "{} all 'contents' must be Form subclasses, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if not isinstance(valid_when, bool):
            raise TypeError(
                "{} 'valid_when' must be bool, not {}".format(
                    type(self).__name__, repr(valid_when)
                )
            )

        self._mask = mask
        self._content = content
        self._valid_when = valid_when
        self._init(parameters=parameters, form_key=form_key)

    @property
    def mask(self):
        return self._mask

    @property
    def content(self):
        return self._content

    @property
    def valid_when(self):
        return self._valid_when

    def copy(
        self,
        mask=UNSET,
        content=UNSET,
        valid_when=UNSET,
        *,
        parameters=UNSET,
        form_key=UNSET,
    ):
        return ByteMaskedForm(
            self._mask if mask is UNSET else mask,
            self._content if content is UNSET else content,
            self._valid_when if valid_when is UNSET else valid_when,
            parameters=self._parameters if parameters is UNSET else parameters,
            form_key=self._form_key if form_key is UNSET else form_key,
        )

    @classmethod
    def simplified(
        cls,
        mask,
        content,
        valid_when,
        *,
        parameters=None,
        form_key=None,
    ):
        if content.is_union:
            return content._union_of_optionarrays("i64", parameters)
        elif content.is_indexed or content.is_option:
            return ak.forms.IndexedOptionForm.simplified(
                "i64",
                content,
                parameters=parameters,
            )
        else:
            return cls(
                mask, content, valid_when, parameters=parameters, form_key=form_key
            )

    @property
    def is_identity_like(self):
        return False

    def __repr__(self):
        args = [
            repr(self._mask),
            repr(self._content),
            repr(self._valid_when),
            *self._repr_args(),
        ]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _to_dict_part(self, verbose, toplevel):
        return self._to_dict_extra(
            {
                "class": "ByteMaskedArray",
                "mask": self._mask,
                "valid_when": self._valid_when,
                "content": self._content._to_dict_part(verbose, toplevel=False),
            },
            verbose,
        )

    @property
    def type(self):
        return ak.types.OptionType(
            self._content.type, parameters=self._parameters
        ).simplify_option_union()

    def __eq__(self, other):
        if isinstance(other, ByteMaskedForm):
            return (
                self._form_key == other._form_key
                and self._mask == other._mask
                and self._valid_when == other._valid_when
                and type_parameters_equal(self._parameters, other._parameters)
                and self._content == other._content
            )
        else:
            return False

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        if self._parameters is not None:
            for key in keys:
                if key in self._parameters:
                    return self._parameters[key]

        return self._content.purelist_parameters(*keys)

    @property
    def purelist_isregular(self):
        return self._content.purelist_isregular

    @property
    def purelist_depth(self):
        return self._content.purelist_depth

    @property
    def minmax_depth(self):
        return self._content.minmax_depth

    @property
    def branch_depth(self):
        return self._content.branch_depth

    @property
    def fields(self):
        return self._content.fields

    @property
    def is_tuple(self):
        return self._content.is_tuple

    @property
    def dimension_optiontype(self):
        return True

    def _columns(self, path, output, list_indicator):
        self._content._columns(path, output, list_indicator)

    def _prune_columns(self, is_inside_record_or_union: bool) -> Self | None:
        next_content = self._content._prune_columns(is_inside_record_or_union)
        if next_content is None:
            return None
        else:
            return self.copy(content=next_content)

    def _select_columns(self, match_specifier: _SpecifierMatcher) -> Self:
        return self.copy(content=self._content._select_columns(match_specifier))

    def _column_types(self):
        return self._content._column_types()

    def __setstate__(self, state):
        if isinstance(state, dict):
            # read data pickled in Awkward 2.x
            self.__dict__.update(state)
        else:
            # read data pickled in Awkward 1.x

            # https://github.com/scikit-hep/awkward/blob/main-v1/src/python/forms.cpp#L206-L213
            has_identities, parameters, form_key, mask, content, valid_when = state

            if form_key is not None:
                form_key = "part0-" + form_key  # only the first partition

            self.__init__(
                mask, content, valid_when, parameters=parameters, form_key=form_key
            )

    def _expected_from_buffers(
        self, getkey: Callable[[Form, str], str], recursive: bool
    ) -> Iterator[tuple[str, DType]]:
        yield (getkey(self, "mask"), index_to_dtype[self._mask])
        if recursive:
            yield from self._content._expected_from_buffers(getkey, recursive)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import multiprocessing
import os
import pickle
import sys
from concurrent.futures import ProcessPoolExecutor

if sys.version_info < (3, 12):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata

import pytest

import awkward as ak


def has_entry_point():
    return bool(importlib_metadata.entry_points(group="awkward.pickle.reduce").names)


pytestmark = pytest.mark.skipif(
    has_entry_point(),
    reason="Custom pickler is already registered!",
)


def _init_process_with_pickler(pickler_source: str, tmp_path):
    # Create custom plugin
    (tmp_path / "impl_pickler.py").write_bytes(pickler_source.encode("UTF-8"))
    dist_info = tmp_path / "impl_pickler-0.0.0.dist-info"
    dist_info.mkdir()
    (dist_info / "entry_points.txt").write_bytes(
        b"[awkward.pickle.reduce]\nimpl = impl_pickler:plugin\n"
    )
    sys.path.insert(0, os.fsdecode(tmp_path))


def _pickle_complex_array_and_return_form_impl():
    array = ak.Array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])[[0, 2]]
    return pickle.loads(pickle.dumps(array)).layout.form


def pickle_complex_array_and_return_form(pickler_source, tmp_path):
    """Create a new (spawned) process, and register the given pickler source
    via entrypoints"""
    with ProcessPoolExecutor(
        1,
        initializer=_init_process_with_pickler,
        initargs=(pickler_source, tmp_path),
        # Don't fork the current process with all of its state
        mp_context=multiprocessing.get_context("spawn"),
    ) as executor:
        pickle_future = executor.submit(_pickle_complex_array_and_return_form_impl)
        return pickle_future.result()


def test_default_pickler():
    assert _pickle_complex_array_and_return_form_impl() == ak.forms.from_dict(
        {"class": "ListOffsetArray", "offsets": "i64", "content": "int64"}
    )


def test_noop_pickler(tmp_path):
    assert pickle_complex_array_and_return_form(
        """
def plugin(obj, protocol: int):
    return NotImplemented""",
        tmp_path,
    ) == ak.forms.from_dict(
        {"class": "ListOffsetArray", "offsets": "i64", "content": "int64"}
    )


def test_non_packing_pickler(tmp_path):
    assert pickle_complex_array_and_return_form(
        """
def plugin(obj, protocol):
    import awkward as ak
    if isinstance(obj, ak.Array):
        form, length, container = ak.to_buffers(obj)
        return (
            object.__new__,
            (ak.Array,),
            (form.to_dict(), length, container, obj.behavior),
        )
    else:
        return NotImplemented""",
        tmp_path,
    ) == ak.forms.from_dict(
        {"class": "ListArray", "starts": "i64", "stops": "i64", "content": "int64"}
    )


def test_malformed_pickler(tmp_path):
    with pytest.raises(RuntimeError, match=r"malformed pickler!"):
        pickle_complex_array_and_return_form(
            """
def plugin(obj, protocol: int):
    raise RuntimeError('malformed pickler!')""",
            tmp_path,
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._backends.typetracer import TypeTracerBackend
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext, ensure_same_backend

__all__ = ("index_in",)

typetracer = TypeTracerBackend.instance()


@high_level_function(module="ak.str")
def index_in(
    array, value_set, *, skip_nones=False, highlevel=True, behavior=None, attrs=None
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        value_set: Array-like data (anything #ak.to_layout recognizes), set of
            values to search for in `array`.
        skip_nones (bool): If True, None values in `array` are not matched
            against `value_set`; otherwise, None is considered a legal value.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns the index of the first pattern in `value_set` that each string in
    `array` matches. If the string is not found within `value_set`, then the
    index is set to None.

    Note: this function does not raise an error if the `array` does not
    contain any string or bytestring data.

    Requires the pyarrow library and calls
    [pyarrow.compute.index_in](https://arrow.apache.org/docs/python/generated/pyarrow.compute.index_in.html).
    """
    # Dispatch
    yield (array, value_set)

    # Implementation
    return _impl(array, value_set, skip_nones, highlevel, behavior, attrs)


def _is_maybe_optional_list_of_string(layout):
    if layout.is_list and layout.parameter("__array__") in {"string", "bytestring"}:
        return True
    elif layout.is_option or layout.is_indexed:
        return _is_maybe_optional_list_of_string(layout.content)
    else:
        return False


def _impl(array, value_set, skip_nones, highlevel, behavior, attrs):
    from awkward._connect.pyarrow import import_pyarrow_compute
    from awkward.operations.str import _apply_through_arrow

    pc = import_pyarrow_compute("ak.str.index_in")

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout, value_set_layout = ensure_same_backend(
            ctx.unwrap(array, allow_record=False),
            ctx.unwrap(value_set, allow_record=False),
        )

    if not _is_maybe_optional_list_of_string(value_set_layout):
        raise TypeError("`value_set` must be 1D array of (possibly missing) strings")

    def apply(layout, **kwargs):
        if _is_maybe_optional_list_of_string(layout):
            return _apply_through_arrow(
                pc.index_in,
                layout,
                value_set_layout,
                skip_nulls=skip_nones,
                expect_option_type=True,
                generate_bitmasks=True,
            )

    out = ak._do.recursively_apply(layout, apply)

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("with_name",)

np = NumpyMetadata.instance()


@high_level_function()
def with_name(array, name, *, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        name (str or None): Name to give to the records or tuples; this assigns
            the `"__record__"` parameter. If None, any existing name is unset.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns an #ak.Array or #ak.Record (or low-level equivalent, if
    `highlevel=False`) with a new name. This function does not change the
    array in-place. If the new name is None, then an array without a name is
    returned.

    The records or tuples may be nested within multiple levels of nested lists.
    If records are nested within records, only the outermost are affected.

    Setting the `"__record__"` parameter makes it possible to add behaviors
    to the data; see #ak.Array and #ak.behavior for a more complete
    description.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, name, highlevel, behavior, attrs)


def _impl(array, name, highlevel, behavior, attrs):
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=True, primitive_policy="error")

    def action(layout, **ignore):
        if isinstance(layout, ak.contents.RecordArray):
            return ak.contents.RecordArray(
                layout._contents,
                layout._fields,
                layout._length,
                parameters={**layout.parameters, "__record__": name},
            )
        else:
            return None

    out = ak._do.recursively_apply(layout, action)

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import (
    HighLevelContext,
    maybe_highlevel_to_lowlevel,
)
from awkward._nplikes import ufuncs
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._regularize import regularize_axis

__all__ = ("softmax",)

np = NumpyMetadata.instance()


@high_level_function()
def softmax(
    x,
    axis=None,
    *,
    keepdims=False,
    mask_identity=False,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        x: The data on which to compute the softmax (anything #ak.to_layout recognizes).
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this function decreases the number of
            dimensions by 1; if True, the output values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, the application of this function on
            empty lists results in None (an option type); otherwise, the
            calculation is followed through with the reducers' identities,
            usually resulting in floating-point `nan`.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Computes the softmax in each group of elements from `x` (many
    types supported, including all Awkward Arrays and Records). The grouping
    is performed the same way as for reducers, though this operation is not a
    reducer and has no identity.

    This function has no NumPy equivalent.

    Passing all arguments to the reducers, the softmax is calculated as

        np.exp(x) / ak.sum(np.exp(x))

    See #ak.sum for a complete description of handling nested lists and
    missing values (None) in reducers, and #ak.mean for an example with another
    non-reducer.
    """
    # Dispatch
    yield (x,)

    # Implementation
    return _impl(x, axis, keepdims, mask_identity, highlevel, behavior, attrs)


def _impl(x, axis, keepdims, mask_identity, highlevel, behavior, attrs):
    axis = regularize_axis(axis)

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        x_layout = ctx.unwrap(x, allow_record=False, primitive_policy="error")
    x = ctx.wrap(x_layout)

    with np.errstate(invalid="ignore", divide="ignore"):
        expx = ufuncs.exp(x)
        denom = ak.operations.ak_sum._impl(
            expx,
            axis,
            keepdims,
            mask_identity,
            highlevel=True,
            behavior=ctx.behavior,
            attrs=ctx.attrs,
        )
        return ctx.wrap(
            maybe_highlevel_to_lowlevel(expx / denom),
            highlevel=highlevel,
            allow_other=True,
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._connect.numpy import UNSUPPORTED
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._regularize import regularize_axis

__all__ = ("min", "nanmin")

np = NumpyMetadata.instance()


@high_level_function()
def min(
    array,
    axis=None,
    *,
    keepdims=False,
    initial=None,
    mask_identity=True,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this reducer decreases the number of
            dimensions by 1; if True, the reduced values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        initial (None or number): The maximum value of an output element, as
            an alternative to the numeric type's natural identity (e.g. infinity
            for floating-point types, a maximum integer for integer types).
            If you use `initial`, you might also want `mask_identity=False`.
        mask_identity (bool): If True, reducing over empty lists results in
            None (an option type); otherwise, reducing over empty lists
            results in the operation's identity.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns the minimum value in each group of elements from `array` (many
    types supported, including all Awkward Arrays and Records). The identity
    of minimization is `inf` if floating-point or the largest integer value
    if applied to integers. This identity is usually masked: the minimum of
    an empty list is None, unless `mask_identity=False`.
    This operation is the same as NumPy's
    [amin](https://docs.scipy.org/doc/numpy/reference/generated/numpy.amin.html)
    if all lists at a given dimension have the same length and no None values,
    but it generalizes to cases where they do not.

    See #ak.sum for a more complete description of nested list and missing
    value (None) handling in reducers.

    See also #ak.nanmin.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(
        array,
        axis,
        keepdims,
        initial,
        mask_identity,
        highlevel,
        behavior,
        attrs,
    )


@high_level_function()
def nanmin(
    array,
    axis=None,
    *,
    keepdims=False,
    initial=None,
    mask_identity=True,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this reducer decreases the number of
            dimensions by 1; if True, the reduced values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        initial (None or number): The maximum value of an output element, as
            an alternative to the numeric type's natural identity (e.g. infinity
            for floating-point types, a maximum integer for integer types).
            If you use `initial`, you might also want `mask_identity=False`.
        mask_identity (bool): If True, reducing over empty lists results in
            None (an option type); otherwise, reducing over empty lists
            results in the operation's identity.

    Like #ak.min, but treating NaN ("not a number") values as missing.

    Equivalent to

        ak.min(ak.nan_to_none(array))

    with all other arguments unchanged.

    See also #ak.min.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(
        ak.operations.ak_nan_to_none._impl(array, False, None),
        axis,
        keepdims,
        initial,
        mask_identity,
        highlevel,
        behavior,
        attrs,
    )


def _impl(array, axis, keepdims, initial, mask_identity, highlevel, behavior, attrs):
    axis = regularize_axis(axis)
    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=False, primitive_policy="error")
    reducer = ak._reducers.Min(initial)

    out = ak._do.reduce(
        layout,
        reducer,
        axis=axis,
        mask=mask_identity,
        keepdims=keepdims,
        behavior=ctx.behavior,
    )
    return ctx.wrap(out, highlevel=highlevel, allow_other=True)


@ak._connect.numpy.implements("amin")
def _nep_18_impl_amin(
    a,
    axis=None,
    out=UNSUPPORTED,
    keepdims=False,
    initial=None,
    where=UNSUPPORTED,
):
    return min(a, axis=axis, keepdims=keepdims, initial=initial)


@ak._connect.numpy.implements("nanmin")
def _nep_18_impl_nanmin(
    a,
    axis=None,
    out=UNSUPPORTED,
    keepdims=False,
    initial=None,
    where=UNSUPPORTED,
):
    return nanmin(a, axis=axis, keepdims=keepdims, initial=initial)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np

import awkward as ak


def test_string():
    source = ak.Array(["abc$¢€", "d¢#", "€e¢", "💰💰"])
    result = source.to_numpy(False)
    expected = np.array(["abc$¢€", "d¢#", "€e¢", "💰💰"])
    assert result.dtype == expected.dtype
    np.testing.assert_equal(result, expected)


def test_bytestring():
    source = ak.Array([b"foo", b"bar", b"catastrophic", b"\x03\x07"])
    result = source.to_numpy(False)
    expected = np.array([b"foo", b"bar", b"catastrophic", b"\x03\x07"])
    assert result.dtype == expected.dtype
    np.testing.assert_equal(result, expected)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._dispatch import high_level_function
from awkward._nplikes.numpy_like import NumpyMetadata

__all__ = ("from_arrow_schema",)

np = NumpyMetadata.instance()


@high_level_function()
def from_arrow_schema(schema):
    """
    Args:
        schema (`pyarrow.Schema`): Apache Arrow schema to convert into an Awkward Form.

    Converts an Apache Arrow schema into an Awkward Form.

    See also #ak.to_arrow, #ak.to_arrow_table, #ak.from_arrow, #ak.to_parquet, #ak.from_parquet.
    """
    return _impl(schema)


def _impl(schema):
    import awkward._connect.pyarrow

    return awkward._connect.pyarrow.form_handle_arrow(schema, pass_empty_field=True)



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_1():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_2():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [3, 3, 3]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_3():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_4():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_5():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_6():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_7():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [3, 4, 4]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_8():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 1, 1]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_9():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 3, 3]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_10():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_11():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_12():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [3, 3, 2]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_13():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 1, 2]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_14():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 0, 1]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_15():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_16():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_17():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [3, 2, 3]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_18():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 2, 0]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_19():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 1, 2]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_20():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_21():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_22():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 3, 3, 4, 5, 5, 5, 5, 5, 6, 7, 8, 10, 11], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 2, 2]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_23():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([2, 1, 0, 1, 2, 0, 1, 2, 2, 2, 1, 2, 1, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [2, 2, 2]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_24():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([1, 0, 2, 3, 1, 2, 0, 0, 1, 1, 2, 3, 1, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [1, 1, 1]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))

def test_cuda_awkward_ListOffsetArray64_flatten_offsets_64_25():
    tooffsets = cupy.array([123, 123, 123], dtype=cupy.int64)
    outeroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    outeroffsetslen = 3
    inneroffsets = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    inneroffsetslen = 3
    funcC = cupy_backend['awkward_ListOffsetArray_flatten_offsets', cupy.int64, cupy.int64, cupy.int64]
    funcC(tooffsets, outeroffsets, outeroffsetslen, inneroffsets, inneroffsetslen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tooffsets = [0, 0, 0]
    assert cupy.array_equal(tooffsets[:len(pytest_tooffsets)], cupy.array(pytest_tooffsets))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext

__all__ = ("from_categorical",)


@high_level_function()
def from_categorical(array, *, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    This function replaces categorical data with non-categorical data (by
    removing the label that declares it as such).

    This is a metadata-only operation; the running time does not scale with the
    size of the dataset. (Conversion to categorical is expensive; conversion
    from categorical is cheap.)

    See also #ak.is_categorical, #ak.categories, #ak.str.to_categorical.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, highlevel, behavior, attrs)


def _impl(array, highlevel, behavior, attrs):
    def action(layout, **kwargs):
        if layout.parameter("__array__") == "categorical":
            out = ak.operations.with_parameter(
                layout, "__array__", None, highlevel=False
            )
            return out

        else:
            return None

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array, allow_record=False, primitive_policy="error")
    out = ak._do.recursively_apply(layout, action)
    return ctx.wrap(out, highlevel=highlevel, allow_other=True)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from collections.abc import Callable

import awkward as ak
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._parameters import parameters_union, type_parameters_equal
from awkward._typing import DType, Iterator, JSONSerializable, Self, final
from awkward._util import UNSET
from awkward.forms.form import Form, _SpecifierMatcher, index_to_dtype

__all__ = ("IndexedForm",)

np = NumpyMetadata.instance()


@final
class IndexedForm(Form):
    is_indexed = True

    def __init__(
        self,
        index,
        content=None,
        *,
        parameters=None,
        form_key=None,
    ):
        if not isinstance(index, str):
            raise TypeError(
                f"{type(self).__name__} 'index' must be of type str, not {index!r}"
            )
        if not isinstance(content, Form):
            raise TypeError(
                "{} all 'contents' must be Form subclasses, not {}".format(
                    type(self).__name__, repr(content)
                )
            )

        self._index = index
        self._content = content
        self._init(parameters=parameters, form_key=form_key)

    @property
    def index(self):
        return self._index

    @property
    def content(self):
        return self._content

    def copy(
        self,
        index=UNSET,
        content=UNSET,
        *,
        parameters=UNSET,
        form_key=UNSET,
    ):
        return IndexedForm(
            self._index if index is UNSET else index,
            self._content if content is UNSET else content,
            parameters=self._parameters if parameters is UNSET else parameters,
            form_key=self._form_key if form_key is UNSET else form_key,
        )

    @classmethod
    def simplified(
        cls,
        index,
        content,
        *,
        parameters=None,
        form_key=None,
    ):
        is_cat = parameters is not None and parameters.get("__array__") == "categorical"

        if content.is_union and not is_cat:
            return content.copy(
                parameters=parameters_union(content._parameters, parameters)
            )

        elif content.is_option:
            return ak.forms.IndexedOptionForm.simplified(
                "i64",
                content.content,
                parameters=parameters_union(content._parameters, parameters),
            )

        elif content.is_indexed:
            return IndexedForm(
                "i64",
                content.content,
                parameters=parameters_union(content._parameters, parameters),
            )

        else:
            return cls(index, content, parameters=parameters, form_key=form_key)

    def __repr__(self):
        args = [repr(self._index), repr(self._content), *self._repr_args()]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _to_dict_part(self, verbose, toplevel):
        return self._to_dict_extra(
            {
                "class": "IndexedArray",
                "index": self._index,
                "content": self._content._to_dict_part(verbose, toplevel=False),
            },
            verbose,
        )

    @property
    def type(self):
        out = self._content.type

        if self._parameters is not None:
            if out._parameters is None:
                out._parameters = self._parameters
            else:
                out._parameters = parameters_union(out._parameters, self._parameters)

            if self._parameters.get("__array__") == "categorical":
                if out._parameters is self._parameters:
                    out._parameters = dict(out._parameters)

                # Restore content __array__
                out_array = self._content.parameter("__array__")
                if out_array is not None:
                    out._parameters["__array__"] = out_array
                else:
                    del out._parameters["__array__"]
                out._parameters["__categorical__"] = True

        return out

    def __eq__(self, other):
        if isinstance(other, IndexedForm):
            return (
                self._form_key == other._form_key
                and self._index == other._index
                and type_parameters_equal(self._parameters, other._parameters)
                and self._content == other._content
            )
        else:
            return False

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        if self._parameters is not None:
            for key in keys:
                if key in self._parameters:
                    return self._parameters[key]

        return self._content.purelist_parameters(*keys)

    @property
    def purelist_isregular(self):
        return self._content.purelist_isregular

    @property
    def purelist_depth(self):
        return self._content.purelist_depth

    @property
    def is_identity_like(self):
        return False

    @property
    def minmax_depth(self):
        return self._content.minmax_depth

    @property
    def branch_depth(self):
        return self._content.branch_depth

    @property
    def fields(self):
        return self._content.fields

    @property
    def is_tuple(self):
        return self._content.is_tuple

    @property
    def dimension_optiontype(self):
        return False

    def _columns(self, path, output, list_indicator):
        self._content._columns(path, output, list_indicator)

    def _prune_columns(self, is_inside_record_or_union: bool) -> Self | None:
        next_content = self._content._prune_columns(is_inside_record_or_union)
        if next_content is None:
            return None
        else:
            return self.copy(content=next_content)

    def _select_columns(self, match_specifier: _SpecifierMatcher) -> Self:
        return self.copy(content=self._content._select_columns(match_specifier))

    def _column_types(self):
        return self._content._column_types()

    def __setstate__(self, state):
        if isinstance(state, dict):
            # read data pickled in Awkward 2.x
            self.__dict__.update(state)
        else:
            # read data pickled in Awkward 1.x

            # https://github.com/scikit-hep/awkward/blob/main-v1/src/python/forms.cpp#L279-L285
            has_identities, parameters, form_key, index, content = state

            if form_key is not None:
                form_key = "part0-" + form_key  # only the first partition

            self.__init__(index, content, parameters=parameters, form_key=form_key)

    def _expected_from_buffers(
        self, getkey: Callable[[Form, str], str], recursive: bool
    ) -> Iterator[tuple[str, DType]]:
        yield (getkey(self, "index"), index_to_dtype[self._index])
        if recursive:
            yield from self._content._expected_from_buffers(getkey, recursive)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext

__all__ = ("replace_substring",)


@high_level_function(module="ak.str")
def replace_substring(
    array,
    pattern,
    replacement,
    *,
    max_replacements=None,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        pattern (str): Substring pattern to look for inside input values.
        replacement (str or bytes): What to replace the pattern with.
        max_replacements (None or int): If not None and not -1, limits the
            maximum number of replacements per string/bytestring, counting from
            the left.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Replaces non-overlapping subsequences of any string or bytestring-valued
    data that match a literal `pattern` with `replacement`.

    The `pattern` and `replacement` are scalars; they cannot be different for
    each string/bytestring in the sample.

    Note: this function does not raise an error if the `array` does not
    contain any string or bytestring data.

    Requires the pyarrow library and calls
    [pyarrow.compute.replace_substring](https://arrow.apache.org/docs/python/generated/pyarrow.compute.replace_substring.html)
    or
    [pyarrow.compute.replace_substring](https://arrow.apache.org/docs/python/generated/pyarrow.compute.replace_substring.html)
    on strings and bytestrings, respectively.

    See also: #ak.str.replace_substring_regex.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(
        array, pattern, replacement, max_replacements, highlevel, behavior, attrs
    )


def _impl(array, pattern, replacement, max_replacements, highlevel, behavior, attrs):
    from awkward._connect.pyarrow import import_pyarrow_compute

    pc = import_pyarrow_compute("g")

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array)

    out = ak._do.recursively_apply(
        layout,
        ak.operations.str._get_ufunc_action(
            pc.replace_substring,
            pc.replace_substring,
            pattern,
            replacement,
            max_replacements=max_replacements,
        ),
    )

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import os

import pytest

import awkward as ak

pyarrow = pytest.importorskip("pyarrow")


def test_simple(tmp_path):
    filename = os.path.join(tmp_path, "whatever.feather")

    array = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]])

    ak.to_feather(array, filename)

    array2 = ak.from_feather(filename)

    assert array2.tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]


def test_complex(tmp_path):
    filename = os.path.join(tmp_path, "whatever.feather")

    array = ak.Array(
        [
            [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}, {"x": 3.3, "y": [1, 2, 3]}],
            [],
            [{"x": 4.4, "y": [1, 2, 3, 4]}, {"x": 5.5, "y": [1, 2, 3, 4, 5]}],
        ]
    )

    ak.to_feather(array, filename)

    array2 = ak.from_feather(filename)

    assert array2.tolist() == [
        [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}, {"x": 3.3, "y": [1, 2, 3]}],
        [],
        [{"x": 4.4, "y": [1, 2, 3, 4]}, {"x": 5.5, "y": [1, 2, 3, 4, 5]}],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext

__all__ = ("to_categorical",)


@high_level_function(module="ak.str")
def to_categorical(array, *, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns a dictionary-encoded version of the given array of strings.
    Creates a categorical dataset, which has the following properties:

       * only distinct values (categories) are stored in their entirety,
       * pointers to those distinct values are represented by integers
         (an #ak.contents.IndexedArray or #ak.contents.IndexedOptionArray
         labeled with parameter `"__array__" = "categorical"`.

    This is equivalent to R's "factor", and Pandas's "categorical".
    It differs from generic uses of #ak.contents.IndexedArray and
    #ak.contents.IndexedOptionArray in Awkward Arrays by the guarantee of no
    duplicate categories and the `"categorical"` parameter.


    Unlike Arrow's `dictionary_encode`, this function has no `null_handling`
    argument. This function's behavior is like`null_handling="mask"` (Arrow's default).
    It is not possible to encode null values in Awkward Array, as #ak.contents.IndexedOptionArray
    cannot contain an option type node.

    Note: this function does not raise an error if the `array` does not
    contain any string or bytestring data.

    Requires the pyarrow library and calls
    [pyarrow.compute.dictionary_encode](https://arrow.apache.org/docs/python/generated/pyarrow.compute.dictionary_encode.html).
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, highlevel, behavior, attrs)


def _impl(array, highlevel, behavior, attrs):
    from awkward._connect.pyarrow import import_pyarrow_compute
    from awkward.operations.str import _apply_through_arrow

    pc = import_pyarrow_compute("ak.str.to_categorical")

    def action(layout, **kwargs):
        if layout.is_list and layout.parameter("__array__") in {"string", "bytestring"}:
            return _apply_through_arrow(
                pc.dictionary_encode, layout, expect_option_type=False
            )

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array)

    out = ak._do.recursively_apply(layout, action)

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pickle

import pytest  # noqa: F401

from awkward._nplikes.shape import unknown_length


def test():
    assert pickle.loads(pickle.dumps(unknown_length)) is unknown_length



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    layout = ak.contents.RecordArray(
        [
            ak.contents.ListOffsetArray(
                ak.index.Index64([0, 3]),
                ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.uint16)),
            ),
            ak.contents.ListOffsetArray(
                ak.index.Index64([0, 2]),
                ak.contents.NumpyArray(np.array([4, 5], dtype=np.uint16)),
            ),
        ],
        ["x", "y"],
    )
    sliced = layout[..., np.newaxis]
    assert sliced.is_equal_to(
        ak.contents.RecordArray(
            [
                ak.contents.RegularArray(
                    ak.contents.ListOffsetArray(
                        ak.index.Index64([0, 3]),
                        ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.uint16)),
                    ),
                    1,
                ),
                ak.contents.RegularArray(
                    ak.contents.ListOffsetArray(
                        ak.index.Index64([0, 2]),
                        ak.contents.NumpyArray(np.array([4, 5], dtype=np.uint16)),
                    ),
                    1,
                ),
            ],
            ["x", "y"],
        )
    )



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_BitMaskedArray_to_IndexedOptionArray64_1():
    toindex = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    frombitmask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = True
    lsb_order = True
    funcC = cupy_backend['awkward_BitMaskedArray_to_IndexedOptionArray', cupy.int64, cupy.uint8]
    funcC(toindex, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [0, -1, -1, -1, -1, -1, -1, -1, 8, -1, -1, -1, -1, -1, -1, -1, 16, -1, -1, -1, -1, -1, -1, -1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_BitMaskedArray_to_IndexedOptionArray64_2():
    toindex = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    frombitmask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = False
    lsb_order = False
    funcC = cupy_backend['awkward_BitMaskedArray_to_IndexedOptionArray', cupy.int64, cupy.uint8]
    funcC(toindex, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_BitMaskedArray_to_IndexedOptionArray64_3():
    toindex = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    frombitmask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = True
    lsb_order = False
    funcC = cupy_backend['awkward_BitMaskedArray_to_IndexedOptionArray', cupy.int64, cupy.uint8]
    funcC(toindex, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_BitMaskedArray_to_IndexedOptionArray64_4():
    toindex = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    frombitmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = True
    lsb_order = True
    funcC = cupy_backend['awkward_BitMaskedArray_to_IndexedOptionArray', cupy.int64, cupy.uint8]
    funcC(toindex, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [0, -1, -1, -1, -1, -1, -1, -1, 8, -1, -1, -1, -1, -1, -1, -1, 16, -1, -1, -1, -1, -1, -1, -1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_BitMaskedArray_to_IndexedOptionArray64_5():
    toindex = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    frombitmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.uint8)
    bitmasklength = 3
    validwhen = False
    lsb_order = False
    funcC = cupy_backend['awkward_BitMaskedArray_to_IndexedOptionArray', cupy.int64, cupy.uint8]
    funcC(toindex, frombitmask, bitmasklength, validwhen, lsb_order)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [0, 1, 2, 3, 4, 5, 6, -1, 8, 9, 10, 11, 12, 13, 14, -1, 16, 17, 18, 19, 20, 21, 22, -1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import packaging.version
import pytest

import awkward as ak

numba = pytest.importorskip("numba")

NUMBA_HAS_NEP_50 = packaging.version.parse(
    numba.__version__
) >= packaging.version.Version("0.59.0")


@pytest.mark.skipif(not NUMBA_HAS_NEP_50, reason="Numba does not have NEP-50 support")
def test_numba_ufunc_nep_50():
    raise NotImplementedError

    @numba.vectorize
    def add(x, y):
        return x + y

    array = ak.values_astype([[1, 2, 3], [4]], np.int8)

    # FIXME: what error will Numba throw here for an out-of-bounds integer?
    with pytest.warns(FutureWarning, match=r"not create a writeable array"):
        result = add(array, np.int16(np.iinfo(np.int8).max + 1))

    flattened = ak.to_numpy(ak.flatten(result))
    assert flattened.dtype == np.dtype(np.int8)


@pytest.mark.skipif(NUMBA_HAS_NEP_50, reason="Numba has NEP-50 support")
def test_numba_ufunc_legacy():
    @numba.vectorize
    def add(x, y):
        return x + y

    array = ak.values_astype([[1, 2, 3], [4]], np.int8)
    with pytest.warns(FutureWarning, match=r"not create a writeable array"):
        result = add(array, np.int16(np.iinfo(np.int8).max + 1))

    flattened = ak.to_numpy(ak.flatten(result))
    # Seems like Numba chooses int64 here unless a 32-bit platform
    assert flattened.dtype == np.dtype(np.int32 if ak._util.bits32 else np.int64)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

"""Interface for plugin-configurable pickle __reduce_ex__ implementation"""

from __future__ import annotations

import sys
import threading
import warnings
from collections.abc import Mapping
from contextlib import contextmanager

from awkward._typing import TYPE_CHECKING, Any, JSONMapping, Protocol, runtime_checkable

if sys.version_info < (3, 12):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata


if TYPE_CHECKING:
    from awkward._nplikes.shape import ShapeItem
    from awkward.highlevel import Array, Record


@runtime_checkable
class PickleReducer(Protocol):
    def __call__(self, obj: Any, protocol: int) -> tuple | NotImplemented:
        ...


_register_lock = threading.Lock()
_plugin: PickleReducer | None = None
_is_registered = False


def _load_reduce_plugin():
    best_plugin = None

    for entry_point in importlib_metadata.entry_points(group="awkward.pickle.reduce"):
        plugin = entry_point.load()

        try:
            assert isinstance(plugin, PickleReducer)
        except AssertionError:
            warnings.warn(
                f"Couldn't load `awkward.pickle.reduce` plugin: {entry_point}",
                stacklevel=2,
            )
            continue

        if best_plugin is not None:
            raise RuntimeError(
                "Encountered multiple Awkward pickle reducers under the `awkward.pickle.reduce` entrypoint"
            )
        best_plugin = plugin

    return best_plugin


def get_custom_reducer() -> PickleReducer | None:
    """
    Returns the implementation of a custom __reduce_ex__ function for Awkward
    highlevel objects, or None if none provided
    """
    global _is_registered, _plugin

    with _register_lock:
        if not _is_registered:
            _plugin = _load_reduce_plugin()
            _is_registered = True

        if _plugin is None:
            return None
        else:
            return _plugin


_DISABLE_CUSTOM_REDUCER = False


@contextmanager
def use_builtin_reducer():
    global _DISABLE_CUSTOM_REDUCER
    old_value, _DISABLE_CUSTOM_REDUCER = _DISABLE_CUSTOM_REDUCER, True
    try:
        yield
    finally:
        _DISABLE_CUSTOM_REDUCER = old_value


def custom_reduce(obj, protocol: int) -> tuple | NotImplemented:
    if (plugin := get_custom_reducer()) is None or _DISABLE_CUSTOM_REDUCER:
        return NotImplemented
    else:
        return plugin(obj, protocol)


def unpickle_array_schema_1(
    form_dict: dict,
    length: ShapeItem,
    container: Mapping[str, Any],
    behavior: JSONMapping | None,
    attrs: JSONMapping | None,
) -> Array:
    from awkward.operations.ak_from_buffers import _impl

    return _impl(
        form_dict,
        length,
        container,
        backend="cpu",
        behavior=behavior,
        attrs=attrs,
        highlevel=True,
        buffer_key="{form_key}-{attribute}",
        byteorder="<",
        simplify=False,
    )


def unpickle_record_schema_1(
    form_dict: dict,
    length: ShapeItem,
    container: Mapping[str, Any],
    behavior: JSONMapping | None,
    attrs: JSONMapping | None,
    at: int,
) -> Record:
    from awkward.highlevel import Record
    from awkward.operations.ak_from_buffers import _impl
    from awkward.record import Record as LowLevelRecord

    array_layout = _impl(
        form_dict,
        length,
        container,
        backend="cpu",
        behavior=behavior,
        attrs=attrs,
        highlevel=False,
        buffer_key="{form_key}-{attribute}",
        byteorder="<",
        simplify=False,
    )
    layout = LowLevelRecord(array_layout, at)
    return Record(layout, behavior=behavior, attrs=attrs)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import packaging.version

_has_checked_version = False


def register_and_check():
    global _has_checked_version

    try:
        import cppyy
    except ImportError as err:
        raise ImportError(
            """install the 'cppyy' package with:

pip install cppyy

or

conda install -c conda-forge cppyy

Note that this must be in a different venv or conda environment from ROOT, if you have installed ROOT.
"""
        ) from err

    if not _has_checked_version:
        if packaging.version.Version(cppyy.__version__) < packaging.version.Version(
            "3.1.0"
        ):
            raise ImportError(
                "Awkward Array can only work with cppyy 3.1.0 or later "
                f"(you have version {cppyy.__version__})"
            )
        _has_checked_version = True



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from collections.abc import Callable

import awkward as ak
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._parameters import type_parameters_equal
from awkward._typing import DType, Iterator, JSONSerializable, Self, final
from awkward._util import UNSET
from awkward.forms.form import Form, _SpecifierMatcher, index_to_dtype

__all__ = ("ListForm",)

np = NumpyMetadata.instance()


@final
class ListForm(Form):
    is_list = True

    def __init__(
        self,
        starts,
        stops,
        content,
        *,
        parameters=None,
        form_key=None,
    ):
        if not isinstance(starts, str):
            raise TypeError(
                "{} 'starts' must be of type str, not {}".format(
                    type(self).__name__, repr(starts)
                )
            )
        if not isinstance(stops, str):
            raise TypeError(
                "{} 'starts' must be of type str, not {}".format(
                    type(self).__name__, repr(starts)
                )
            )
        if not isinstance(content, Form):
            raise TypeError(
                "{} all 'contents' must be Form subclasses, not {}".format(
                    type(self).__name__, repr(content)
                )
            )

        self._starts = starts
        self._stops = stops
        self._content = content
        self._init(parameters=parameters, form_key=form_key)

    @property
    def starts(self):
        return self._starts

    @property
    def stops(self):
        return self._stops

    @property
    def content(self):
        return self._content

    def copy(
        self,
        starts=UNSET,
        stops=UNSET,
        content=UNSET,
        *,
        parameters=UNSET,
        form_key=UNSET,
    ):
        return ListForm(
            self._starts if starts is UNSET else starts,
            self._stops if stops is UNSET else stops,
            self._content if content is UNSET else content,
            parameters=self._parameters if parameters is UNSET else parameters,
            form_key=self._form_key if form_key is UNSET else form_key,
        )

    @classmethod
    def simplified(
        cls,
        starts,
        stops,
        content,
        *,
        parameters=None,
        form_key=None,
    ):
        return cls(starts, stops, content, parameters=parameters, form_key=form_key)

    def __repr__(self):
        args = [
            repr(self._starts),
            repr(self._stops),
            repr(self._content),
            *self._repr_args(),
        ]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _to_dict_part(self, verbose, toplevel):
        return self._to_dict_extra(
            {
                "class": "ListArray",
                "starts": self._starts,
                "stops": self._stops,
                "content": self._content._to_dict_part(verbose, toplevel=False),
            },
            verbose,
        )

    @property
    def type(self):
        return ak.types.ListType(
            self._content.type,
            parameters=self._parameters,
        )

    def __eq__(self, other):
        if isinstance(other, ListForm):
            return (
                self._form_key == other._form_key
                and self._starts == other._starts
                and self._stops == other._stops
                and type_parameters_equal(self._parameters, other._parameters)
                and self._content == other._content
            )
        else:
            return False

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        if self._parameters is not None:
            for key in keys:
                if key in self._parameters:
                    return self._parameters[key]

        return self._content.purelist_parameters(*keys)

    @property
    def purelist_isregular(self):
        return False

    @property
    def purelist_depth(self):
        if self.parameter("__array__") in ("string", "bytestring"):
            return 1
        else:
            return self._content.purelist_depth + 1

    @property
    def is_identity_like(self):
        return False

    @property
    def minmax_depth(self):
        if self.parameter("__array__") in ("string", "bytestring"):
            return (1, 1)
        else:
            mindepth, maxdepth = self._content.minmax_depth
            return (mindepth + 1, maxdepth + 1)

    @property
    def branch_depth(self):
        if self.parameter("__array__") in ("string", "bytestring"):
            return (False, 1)
        else:
            branch, depth = self._content.branch_depth
            return (branch, depth + 1)

    @property
    def fields(self):
        return self._content.fields

    @property
    def is_tuple(self):
        return self._content.is_tuple

    @property
    def dimension_optiontype(self):
        return False

    def _columns(self, path, output, list_indicator):
        if (
            self.parameter("__array__") not in ("string", "bytestring")
            and list_indicator is not None
        ):
            path = (*path, list_indicator)
        self._content._columns(path, output, list_indicator)

    def _prune_columns(self, is_inside_record_or_union: bool) -> Self | None:
        next_content = self._content._prune_columns(is_inside_record_or_union)
        if next_content is None:
            return None
        else:
            return self.copy(content=next_content)

    def _select_columns(self, match_specifier: _SpecifierMatcher) -> Self:
        return self.copy(content=self._content._select_columns(match_specifier))

    def _column_types(self):
        if self.parameter("__array__") in ("string", "bytestring"):
            return ("string",)
        else:
            return self._content._column_types()

    def __setstate__(self, state):
        if isinstance(state, dict):
            # read data pickled in Awkward 2.x
            self.__dict__.update(state)
        else:
            # read data pickled in Awkward 1.x

            # https://github.com/scikit-hep/awkward/blob/main-v1/src/python/forms.cpp#L374-L381
            has_identities, parameters, form_key, starts, stops, content = state

            if form_key is not None:
                form_key = "part0-" + form_key  # only the first partition

            self.__init__(
                starts, stops, content, parameters=parameters, form_key=form_key
            )

    def _expected_from_buffers(
        self, getkey: Callable[[Form, str], str], recursive: bool
    ) -> Iterator[tuple[str, DType]]:
        yield (getkey(self, "starts"), index_to_dtype[self._starts])
        yield (getkey(self, "stops"), index_to_dtype[self._stops])
        if recursive:
            yield from self._content._expected_from_buffers(getkey, recursive)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext

__all__ = ("extract_regex",)


@high_level_function(module="ak.str")
def extract_regex(array, pattern, *, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        pattern (str or bytes): Regular expression with named capture fields.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Returns None for every string in `array` if it does not match `pattern`;
    otherwise, a record whose fields are named capture groups and whose
    contents are the substrings they've captured.

    Uses [Google RE2](https://github.com/google/re2/wiki/Syntax), and `pattern` must
    contain named groups. The syntax for a named group is `(?P<...>...)` in which
    the first `...` is a name and the last `...` is a regular expression.

    For example,

        >>> array = ak.Array([["one1", "two2", "three3"], [], ["four4", "five5"]])
        >>> result = ak.str.extract_regex(array, "(?P<vowel>[aeiou])(?P<number>[0-9]+)")
        >>> result.show(type=True)
        type: 3 * var * ?{
            vowel: ?string,
            number: ?string
        }
        [[{vowel: 'e', number: '1'}, {vowel: 'o', number: '2'}, {vowel: 'e', number: '3'}],
         [],
         [None, {vowel: 'e', number: '5'}]]

    (The string `"four4"` does not match because the vowel is not immediately before
    the number.)

    Regular expressions with unnamed groups or features not implemented by RE2 raise an error.

    Note: this function does not raise an error if the `array` does not
    contain any string or bytestring data.

    Requires the pyarrow library and calls
    [pyarrow.compute.extract_regex](https://arrow.apache.org/docs/python/generated/pyarrow.compute.extract_regex.html).
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, pattern, highlevel, behavior, attrs)


def _impl(array, pattern, highlevel, behavior, attrs):
    from awkward._connect.pyarrow import import_pyarrow_compute

    pc = import_pyarrow_compute("x")

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout = ctx.unwrap(array)

    out = ak._do.recursively_apply(
        layout,
        ak.operations.str._get_ufunc_action(
            pc.extract_regex,
            pc.extract_regex,
            pattern,
            generate_bitmasks=True,
            bytestring_to_string=False,
            expect_option_type=True,
        ),
    )

    return ctx.wrap(out, highlevel=highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import datetime

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    array = ak.Array(
        [
            [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
            [],
            [
                {"x": 3.3, "y": [1, 2, None, 3]},
                False,
                False,
                True,
                {"x": 4.4, "y": [1, 2, None, 3, 4]},
            ],
        ]
    )

    assert ak.operations.full_like(array, 12.3).to_list() == [
        [{"x": 12.3, "y": []}, {"x": 12.3, "y": [12]}, {"x": 12.3, "y": [12, 12]}],
        [],
        [
            {"x": 12.3, "y": [12, 12, None, 12]},
            True,
            True,
            True,
            {"x": 12.3, "y": [12, 12, None, 12, 12]},
        ],
    ]

    assert ak.operations.zeros_like(array).to_list() == [
        [{"x": 0.0, "y": []}, {"x": 0.0, "y": [0]}, {"x": 0.0, "y": [0, 0]}],
        [],
        [
            {"x": 0.0, "y": [0, 0, None, 0]},
            False,
            False,
            False,
            {"x": 0.0, "y": [0, 0, None, 0, 0]},
        ],
    ]

    assert ak.operations.ones_like(array).to_list() == [
        [{"x": 1.0, "y": []}, {"x": 1.0, "y": [1]}, {"x": 1.0, "y": [1, 1]}],
        [],
        [
            {"x": 1.0, "y": [1, 1, None, 1]},
            True,
            True,
            True,
            {"x": 1.0, "y": [1, 1, None, 1, 1]},
        ],
    ]

    array = ak.Array([["one", "two", "three"], [], ["four", "five"]])
    assert ak.operations.full_like(array, "hello").to_list() == [
        ["hello", "hello", "hello"],
        [],
        ["hello", "hello"],
    ]
    assert ak.operations.full_like(array, 1).to_list() == [
        ["1", "1", "1"],
        [],
        ["1", "1"],
    ]
    assert ak.operations.full_like(array, 0).to_list() == [
        ["0", "0", "0"],
        [],
        ["0", "0"],
    ]
    assert ak.operations.ones_like(array).to_list() == [
        ["1", "1", "1"],
        [],
        ["1", "1"],
    ]
    assert ak.operations.zeros_like(array).to_list() == [
        ["", "", ""],
        [],
        ["", ""],
    ]

    array = ak.Array([[b"one", b"two", b"three"], [], [b"four", b"five"]])
    assert ak.operations.full_like(array, b"hello").to_list() == [
        [b"hello", b"hello", b"hello"],
        [],
        [b"hello", b"hello"],
    ]
    assert ak.operations.full_like(array, 1).to_list() == [
        [b"1", b"1", b"1"],
        [],
        [b"1", b"1"],
    ]
    assert ak.operations.full_like(array, 0).to_list() == [
        [b"0", b"0", b"0"],
        [],
        [b"0", b"0"],
    ]
    assert ak.operations.ones_like(array).to_list() == [
        [b"1", b"1", b"1"],
        [],
        [b"1", b"1"],
    ]
    assert ak.operations.zeros_like(array).to_list() == [
        [b"", b"", b""],
        [],
        [b"", b""],
    ]


def test_full_like_types():
    array = ak.highlevel.Array(
        np.array(["2020-07-27T10:41:11", "2019-01-01", "2020-01-01"], "datetime64[s]")
    )

    assert ak.operations.full_like(array, "2020-07-27T10:41:11").to_list() == [
        datetime.datetime(2020, 7, 27, 10, 41, 11),
        datetime.datetime(2020, 7, 27, 10, 41, 11),
        datetime.datetime(2020, 7, 27, 10, 41, 11),
    ]

    array = np.array(
        ["2020-07-27T10:41:11", "2019-01-01", "2020-01-01"], "datetime64[25s]"
    )

    assert ak.operations.full_like(array, "2021-06-03T10:00").to_list() == [
        datetime.datetime(2021, 6, 3, 10, 0),
        datetime.datetime(2021, 6, 3, 10, 0),
        datetime.datetime(2021, 6, 3, 10, 0),
    ]

    array = ak.contents.NumpyArray(np.array([0, 2, 2, 3], dtype="i4"))

    assert str(ak.operations.full_like(array, 11, dtype="i8").type) == "4 * int64"
    assert (
        str(ak.operations.full_like(array, 11, dtype=np.dtype(np.int64)).type)
        == "4 * int64"
    )
    assert str(ak.operations.full_like(array, 11, dtype=np.int64).type) == "4 * int64"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test():
    assert ak.to_numpy(ak.Array([{}, {}])).tolist() == [(), ()]
